#!/usr/bin/env python3

# Copyright (c) 2016, Suprock Technologies
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
# OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os.path


def output_shift(f, input_index, left_shift_count, mask=0xFF, t="uint32_t"):
    if mask == 0xFF:
        input_str = "input[{}]".format(input_index)
    else:
        input_str = "(input[{}] & 0x{:02x})".format(input_index, mask)

    if left_shift_count == 0:
        f.write("(({}){})".format(t, input_str))
    elif left_shift_count > 0:
        f.write("(({}){} << {})".format(t, input_str, left_shift_count))
    else:
        f.write("(({}){} >> {})".format(t, input_str, -left_shift_count))


def generate_single_unpack(f, packsize, offset, output_index, output_type):
    """
    Generate instructions for unpacking a single value.
    """
    if packsize == 0:
        f.write("\toutput[{}] = ({})0;\n".format(output_index, output_type))
        return

    f.write("\t// read output[{}]: from bit {} to {}\n".format(
            output_index, offset, offset + packsize - 1))
    f.write("\tvalue = ")

    remaining = packsize
    collected = 0
    while remaining + (offset % 8) >= 8:
        if offset % 8 == 0:
            # use all bits of leading byte
            collected += 8
            output_shift(f, offset // 8, 32 - collected)
            offset += 8
            remaining -= 8
        else:
            # use partial bits of leading byte
            to_collect = 8 - (offset % 8)
            mask = 0xFF >> (8 - to_collect)
            collected += to_collect
            output_shift(f, offset // 8, 32 - collected, mask)
            offset += to_collect
            remaining -= to_collect

        if remaining > 0:
            f.write(" |\n\t\t\t")

    if remaining > 0:
        # collect trailing bits.
        mask = ((0xFF << (8 - remaining)) & 0xFF) >> (offset % 8)
        collected += 8 - (offset % 8)  # we're collecting more than remaining
        output_shift(f, offset // 8, 32 - collected, mask)

    f.write(";\n")
    right_shift_count = 32 - packsize
    if right_shift_count != 0:
        f.write("\toutput[{}] = ({})(value >> {});\n".format(
                output_index, output_type, right_shift_count))
    else:
        f.write("\toutput[{}] = ({})value;\n".format(
                output_index, output_type))


def generate_unpack_function(f, func_name, count, signed, packsize, offset,
                             output_type):
    """
    Generate instructions for unpacking a whole array of values.
    """

    signed_str = "signed" if signed else "unsigned"
    temp_type = "int32_t" if signed else "uint32_t"

    s = "static void {}(const uint8_t * input, {} * output) {{\n"
    f.write(s.format(func_name, output_type))
    f.write("\t// Unpack {} values into output\n".format(count))
    f.write("\t// Values are {}-bit {}\n".format(packsize, signed_str))
    f.write("\t// Skip {} bits, read {} bits of data.\n".format(
            offset, packsize * count))
    f.write("\t// Input must be at least {} bytes long\n".format(
            (packsize * count + offset + 7) // 8))
    f.write("\n")
    if packsize == 0:
        f.write("\t(void)input; // suppress unused parameter warning\n")
    else:
        f.write("\t{} value;\n".format(temp_type))

    for i in range(count):
        f.write("\n")
        generate_single_unpack(f, packsize, offset, i, output_type)
        offset += packsize

    f.write("}\n")


def generate_c_file(f, header_name, packsize, maxcount):
    f.write("// AUTOGENERATED FILE. DO NOT MODIFY.\n")
    f.write("\n")
    f.write("#include <stdlib.h>\n")
    f.write("#include <stdint.h>\n")
    f.write("\n")
    f.write('#include "{}"\n'.format(header_name))
    f.write("\n")
    f.write("\n")

    output_type = "double"

    for count in range(1, maxcount + 1):
        if packsize == 0:
            func_name = "unpack_{}_{}bit".format(count, packsize)
            generate_unpack_function(f, func_name, count, False, packsize, 0,
                                     output_type)
        else:
            for offset in range(8):
                for signed in [True, False]:
                    if packsize == 1 and signed:
                        continue
                    f.write("\n")
                    signed_str = "signed" if signed else "unsigned"
                    func_name = "unpack_{}_{}bit_{}_{}off".format(count, packsize,
                                                                  signed_str,
                                                                  offset)
                    generate_unpack_function(f, func_name, count, signed, packsize,
                                             offset, output_type)

    f.write("\n")
    f.write("\n")
    s = "unpack_func_t unpack_{}bit[UNPACK_{}BIT_MAX_COUNT][8][2] = {{\n"
    f.write(s.format(packsize, packsize))

    for count in range(1, maxcount + 1):
        f.write("\t{\n")
        for offset in range(8):
            f.write("\t\t{\n")
            for signed in [False, True]:
                signed_str = "signed" if signed else "unsigned"
                if packsize == 0:
                    func_name = "unpack_{}_{}bit".format(count, packsize)
                elif packsize == 1 and signed:
                    func_name = "NULL"
                else:
                    func_name = "unpack_{}_{}bit_{}_{}off".format(count, packsize,
                                                                  signed_str,
                                                                  offset)
                f.write("\t\t\t{},\n".format(func_name))
            f.write("\t\t},\n")
        f.write("\t},\n")
    f.write("};\n")


def generate_h_file(f, header_name, packsize, maxcount):
    header_define = header_name.upper().replace(".", "_") + "_"

    f.write("// AUTOGENERATED FILE. DO NOT MODIFY.\n")
    f.write("\n")
    f.write("#ifndef {}\n".format(header_define))
    f.write("#define {}\n".format(header_define))
    f.write("\n")
    f.write("\n")
    f.write('#include "unpack.h"\n')
    f.write("\n")
    f.write("\n")
    f.write("#define UNPACK_{}BIT_MAX_COUNT {}\n".format(packsize, maxcount))
    f.write("\n")
    s = "extern unpack_func_t unpack_{}bit[UNPACK_{}BIT_MAX_COUNT][8][2];"
    f.write(s.format(packsize, packsize))
    f.write("\n")
    f.write("\n")
    f.write("#endif /* {} */\n".format(header_define))


def generate_main_h(f):
    f.write("// AUTOGENERATED FILE. DO NOT MODIFY.\n")
    f.write("\n")
    f.write("#ifndef UNPACK_H_\n")
    f.write("#define UNPACK_H_\n")
    f.write("\n")
    f.write("\n")
    f.write('#include <stdint.h>\n')
    f.write("\n")
    f.write("\n")
    f.write("#ifdef __cplusplus\n")
    f.write('extern "C" {\n')
    f.write("#endif\n")
    f.write("\n")
    f.write("\n")
    f.write("// cross platform definitions for shared/static library support\n")
    f.write("// see https://gcc.gnu.org/wiki/Visibility\n")
    f.write("#if defined _WIN32 || defined __CYGWIN__\n")
    f.write("#  define UNPACK_HELPER_DLL_IMPORT __declspec(dllimport)\n")
    f.write("#  define UNPACK_HELPER_DLL_EXPORT __declspec(dllexport)\n")
    f.write("#  define UNPACK_HELPER_DLL_LOCAL\n")
    f.write("#else\n")
    f.write("#  if __GNUC__ >= 4\n")
    f.write('#    define UNPACK_HELPER_DLL_IMPORT __attribute__ ((visibility ("default")))\n')
    f.write('#    define UNPACK_HELPER_DLL_EXPORT __attribute__ ((visibility ("default")))\n')
    f.write('#    define UNPACK_HELPER_DLL_LOCAL  __attribute__ ((visibility ("hidden")))\n')
    f.write("#  else\n")
    f.write("#    define UNPACK_HELPER_DLL_IMPORT\n")
    f.write("#    define UNPACK_HELPER_DLL_EXPORT\n")
    f.write("#    define UNPACK_HELPER_DLL_LOCAL\n")
    f.write("#  endif\n")
    f.write("#endif\n")
    f.write("\n")
    f.write("#ifdef UNPACK_STATIC_LIB // defined if unpack library is compiled as static lib\n")
    f.write("#  define UNPACK_API\n")
    f.write("#  define UNPACK_LOCAL\n")
    f.write("#else // UNPACK_STATIC_LIB is not defined: this means unpack is a DLL\n")
    f.write("#  ifdef UNPACK_API_EXPORTS // defined if we are building the library (instead of using it)\n")
    f.write("#    define UNPACK_API UNPACK_HELPER_DLL_EXPORT\n")
    f.write("#  else\n")
    f.write("#    define UNPACK_API UNPACK_HELPER_DLL_IMPORT\n")
    f.write("#  endif // UNPACK_API_EXPORTS\n")
    f.write("#  define UNPACK_LOCAL UNPACK_HELPER_DLL_LOCAL\n")
    f.write("#endif // UNPACK_STATIC_LIB\n")
    f.write("\n")
    f.write("\n")
    f.write("typedef void (*unpack_func_t)(const uint8_t *, double *);\n")
    f.write("typedef uint64_t (*unwrap_func_t)(const uint8_t *, uint64_t);\n")
    f.write("typedef uint8_t (*unpack_id_func_t)(const uint8_t *);\n")
    f.write("\n")
    f.write("UNPACK_API unpack_func_t find_unpack(int count, int bits, int is_signed, int offset);\n")
    f.write("UNPACK_API unwrap_func_t find_unwrap(int bits, int offset);\n")
    f.write("UNPACK_API unpack_id_func_t find_unpack_id(int bits, int offset);\n")
    f.write("\n")
    f.write("\n")
    f.write("#ifdef __cplusplus\n")
    f.write("}\n")
    f.write("#endif\n")
    f.write("\n")
    f.write("\n")
    f.write("#endif /* UNPACK_H_ */\n")


def generate_main_c(f, sizes):
    f.write("// AUTOGENERATED FILE. DO NOT MODIFY.\n")
    f.write("\n")
    f.write("#include <stdlib.h>\n")
    f.write('#include <stdint.h>\n')
    f.write("\n")
    f.write('#include "unpack.h"\n')
    f.write('#include "unpack_all_sizes.h"\n')
    f.write('#include "unwrap.h"\n')
    f.write('#include "unpack_id.h"\n')
    f.write("\n")
    f.write("\n")
    f.write("UNPACK_API unpack_func_t find_unpack(int count, int bits, int is_signed, int offset) {\n")
    f.write("\tif (0 <= offset && offset <= 7) {\n")
    f.write("\t\tswitch (bits) {\n")

    for packsize in sizes:
        f.write("\t\t\tcase {}:\n".format(packsize))
        f.write("\t\t\t\tif (1 <= count && count <= UNPACK_{}BIT_MAX_COUNT) {{\n".format(packsize))
        f.write("\t\t\t\t\treturn unpack_{}bit[count - 1][offset][is_signed != 0];\n".format(packsize))
        f.write("\t\t\t\t}\n")
        f.write("\t\t\t\tbreak;\n")

    f.write("\t\t}\n")
    f.write("\t}\n")
    f.write("\n")
    f.write("\treturn NULL;\n")
    f.write("}\n")
    f.write("\n")
    f.write("UNPACK_API unwrap_func_t find_unwrap(int bits, int offset) {\n")
    f.write("\tif (0 <= offset && offset <= 7) {\n")
    f.write("\t\tif (UNWRAP_MIN_COUNT <= bits && bits <= UNWRAP_MAX_COUNT) {\n")
    f.write("\t\t\treturn unwrap[bits - UNWRAP_MIN_COUNT][offset];\n")
    f.write("\t\t}\n")
    f.write("\t}\n")
    f.write("\n")
    f.write("\treturn NULL;\n")
    f.write("}\n")
    f.write("\n")
    f.write("UNPACK_API unpack_id_func_t find_unpack_id(int bits, int offset) {\n")
    f.write("\tif (0 <= offset && offset <= 7) {\n")
    f.write("\t\tif (UNPACK_ID_MIN_COUNT <= bits && bits <= UNPACK_ID_MAX_COUNT) {\n")
    f.write("\t\t\treturn unpack_id[bits - UNPACK_ID_MIN_COUNT][offset];\n")
    f.write("\t\t}\n")
    f.write("\t}\n")
    f.write("\n")
    f.write("\treturn NULL;\n")
    f.write("}\n")


def generate_all_sizes_h(f, sizes):
    f.write("// AUTOGENERATED FILE. DO NOT MODIFY.\n")
    f.write("\n")
    f.write("#ifndef UNPACK_ALL_SIZES_H_\n")
    f.write("#define UNPACK_ALL_SIZES_H_\n")
    f.write("\n")
    f.write("\n")
    for packsize in sizes:
        f.write('#include "unpack{}.h"\n'.format(packsize))
    f.write("\n")
    f.write("\n")
    f.write("#endif /* UNPACK_ALL_SIZES_H_ */\n")


def generate_unwrap_counter_function(f, func_name, counter_size, offset):
    """
    Generate instructions for a single counter size.
    """

    f.write("static uint64_t {}(const uint8_t * input, uint64_t last) {{\n".format(
            func_name))
    f.write("\t// Unpack a single value\n".format(counter_size))
    f.write("\t// Skip {} bit, read {} bits of data\n".format(offset,
                                                              counter_size))
    f.write("\n")
    f.write("\tuint64_t new_counter = ")

    remaining = counter_size
    collected = 0
    while remaining >= 8:
        if offset % 8 == 0:
            # use all bits of leading byte
            collected += 8
            output_shift(f, offset // 8, counter_size - collected)
            offset += 8
            remaining -= 8
        else:
            # use partial bits of leading byte
            to_collect = 8 - (offset % 8)
            mask = 0xFF >> (8 - to_collect)
            collected += to_collect
            output_shift(f, offset // 8, counter_size - collected, mask)
            offset += to_collect
            remaining -= to_collect

        if remaining > 0:
            f.write(" |\n\t\t\t")

    if remaining > 0:
        # collect trailing bits. Offset is a multiple of 8 at this point
        # no need for a mask, the unneeded bit will be right shifted out
        collected += 8  # we're collecting more than remaining
        output_shift(f, offset // 8, counter_size - collected)

    f.write(";\n")
    f.write("\n")
    f.write("\tuint64_t mask = ((uint64_t)1 << {}) - 1;\n".format(
            counter_size))
    f.write("\n")
    f.write("\t// do the unwrapping\n")
    f.write("\tif (new_counter <= (last & mask)) {\n")
    f.write('\t\t// set the "carry" bit\n')
    f.write("\t\tnew_counter |= (uint64_t)1 << {};\n".format(counter_size))
    f.write("\t}\n")
    f.write("\n")
    f.write("\t// add higher bits\n")
    f.write("\tnew_counter += last & (~mask);\n")
    f.write("\n")
    f.write("\treturn new_counter;\n")
    f.write("}\n")


def generate_unwrap_counter_c(f, counter_sizes, header_name):
    f.write("// AUTOGENERATED FILE. DO NOT MODIFY.\n")
    f.write("\n")
    f.write("#include <stdint.h>")
    f.write("\n")
    f.write('#include "{}"'.format(header_name))
    f.write("\n")

    for counter_size in counter_sizes:
        for offset in range(8):
            f.write("\n")
            func_name = "unwrap_{}bit_counter_{}off".format(counter_size,
                                                            offset)
            generate_unwrap_counter_function(f, func_name, counter_size,
                                             offset)

    f.write("\n")
    f.write("\n")
    f.write("unwrap_func_t unwrap[UNWRAP_MAX_COUNT - UNWRAP_MIN_COUNT + 1][8] = {\n")

    for counter_size in counter_sizes:
        f.write("\t{\n")
        for offset in range(8):
            func_name = "unwrap_{}bit_counter_{}off".format(counter_size,
                                                            offset)
            f.write("\t\t{},\n".format(func_name))
        f.write("\t},\n")
    f.write("};\n")


def generate_unwrap_counter_h(f, counter_sizes, header_name):
    header_define = header_name.upper().replace(".", "_") + "_"

    f.write("// AUTOGENERATED FILE. DO NOT MODIFY.\n")
    f.write("\n")
    f.write("#ifndef {}\n".format(header_define))
    f.write("#define {}\n".format(header_define))
    f.write("\n")
    f.write("\n")
    f.write('#include "unpack.h"\n')
    f.write("\n")
    f.write("\n")
    f.write("#define UNWRAP_MIN_COUNT {}\n".format(counter_sizes[0]))
    f.write("#define UNWRAP_MAX_COUNT {}\n".format(counter_sizes[-1]))
    f.write("\n")
    f.write("extern unwrap_func_t unwrap[UNWRAP_MAX_COUNT - UNWRAP_MIN_COUNT + 1][8];\n")
    f.write("\n")
    f.write("\n")
    f.write("#endif /* {} */\n".format(header_define))


def generate_unpack_id_function(f, func_name, id_size, offset):
    """
    Generate instructions for a single id size.
    """

    f.write("static uint8_t {}(const uint8_t * input) {{\n".format(
            func_name))
    f.write("\t// Unpack a single value\n")
    f.write("\t// Skip {} bit, read {} bits of data\n".format(offset,
                                                              id_size))
    f.write("\n")
    f.write("\tuint8_t id = ")

    if id_size == 0:
        f.write("0")
    else:
        # create a 16 bit mask representing the bits to use
        mask_shift = (16 - id_size - offset)
        mask = ((1 << id_size) - 1) << mask_shift
        output_shift(f, 0, 8 - mask_shift, mask >> 8, "uint8_t")
        if mask & 0xFF != 0:
            f.write(" |\n\t\t\t")
            output_shift(f, 1, -mask_shift, mask & 0xFF, "uint8_t")

    f.write(";\n")
    f.write("\n")

    if id_size == 0:
        f.write("\t(void)input; // suppress unused parameter warning\n")
        f.write("\n")

    f.write("\treturn id;\n")
    f.write("}\n")


def generate_unpack_id_c(f, id_sizes, header_name):
    f.write("// AUTOGENERATED FILE. DO NOT MODIFY.\n")
    f.write("\n")
    f.write("#include <stdint.h>")
    f.write("\n")
    f.write('#include "{}"'.format(header_name))
    f.write("\n")

    for id_size in id_sizes:
        if id_size == 0:
            f.write("\n")
            func_name = "unpack_{}bit_id".format(id_size)
            generate_unpack_id_function(f, func_name, id_size, 0)
            continue
        for offset in range(8):
            f.write("\n")
            func_name = "unpack_{}bit_id_{}off".format(id_size, offset)
            generate_unpack_id_function(f, func_name, id_size, offset)

    f.write("\n")
    f.write("\n")
    f.write("unpack_id_func_t unpack_id[UNPACK_ID_MAX_COUNT - UNPACK_ID_MIN_COUNT + 1][8] = {\n")

    for id_size in id_sizes:
        f.write("\t{\n")
        for offset in range(8):
            if id_size == 0:
                func_name = "unpack_{}bit_id".format(id_size)
            else:
                func_name = "unpack_{}bit_id_{}off".format(id_size, offset)
            f.write("\t\t{},\n".format(func_name))
        f.write("\t},\n")
    f.write("};\n")


def generate_unpack_id_h(f, counter_sizes, header_name):
    header_define = header_name.upper().replace(".", "_") + "_"

    f.write("// AUTOGENERATED FILE. DO NOT MODIFY.\n")
    f.write("\n")
    f.write("#ifndef {}\n".format(header_define))
    f.write("#define {}\n".format(header_define))
    f.write("\n")
    f.write("\n")
    f.write('#include "unpack.h"\n')
    f.write("\n")
    f.write("\n")
    f.write("#define UNPACK_ID_MIN_COUNT {}\n".format(counter_sizes[0]))
    f.write("#define UNPACK_ID_MAX_COUNT {}\n".format(counter_sizes[-1]))
    f.write("\n")
    f.write("extern unpack_id_func_t unpack_id[UNPACK_ID_MAX_COUNT - UNPACK_ID_MIN_COUNT + 1][8];\n")
    f.write("\n")
    f.write("\n")
    f.write("#endif /* {} */\n".format(header_define))


def generate_files(path):
    sizes = list(range(0, 32 + 1))

    with open(os.path.join(path, "inc/unpack.h"), "w") as main_h_file:
        generate_main_h(main_h_file)
    with open(os.path.join(path, "src/unpack.c"), "w") as main_c_file:
        generate_main_c(main_c_file, sizes)

    for packsize in sizes:
        if packsize in [16, 24, 32]:
            maxcount = 128 * 8 // packsize
        elif packsize >= 8:
            maxcount = 64 * 8 // packsize
        else:
            maxcount = 8

        h_name = "unpack{}.h".format(packsize)
        c_name = "unpack{}.c".format(packsize)

        with open(os.path.join(path, "src", h_name), "w") as h_file:
            generate_h_file(h_file, h_name, packsize, maxcount)
        with open(os.path.join(path, "src", c_name), "w") as c_file:
            generate_c_file(c_file, h_name, packsize, maxcount)

    with open(os.path.join(path, "src/unpack_all_sizes.h"), "w") as all_h_file:
        generate_all_sizes_h(all_h_file, sizes)

    counter_sizes = list(range(8, 32 + 1))
    unwrap_header_name = "unwrap.h"
    with open(os.path.join(path, "src", unwrap_header_name), "w") as unwrap_h_file:
        generate_unwrap_counter_h(unwrap_h_file, counter_sizes,
                                  unwrap_header_name)
    with open(os.path.join(path, "src", "unwrap.c"), "w") as unwrap_c_file:
        generate_unwrap_counter_c(unwrap_c_file, counter_sizes,
                                  unwrap_header_name)

    id_sizes = list(range(0, 8 + 1))
    unpack_id_header_name = "unpack_id.h"
    with open(os.path.join(path, "src", unpack_id_header_name), "w") as unpack_id_h_file:
        generate_unpack_id_h(unpack_id_h_file, id_sizes, unpack_id_header_name)
    with open(os.path.join(path, "src", "unpack_id.c"), "w") as unpack_id_c_file:
        generate_unpack_id_c(unpack_id_c_file, id_sizes, unpack_id_header_name)

if __name__ == "__main__":
    generate_files(".")



import math
import struct
import time

import asphodel


def find_usb_devices():
    devices_by_serial = {}
    for device in asphodel.find_usb_devices():
        try:
            device.open()
            serial_number = device.get_serial_number()
            devices_by_serial[serial_number] = device
        except asphodel.AsphodelError:
            continue
        finally:
            device.close()
    return devices_by_serial


def find_tcp_devices():
    devices_by_serial = {}
    for device in asphodel.find_tcp_devices():
        adv = device.tcp_get_advertisement()
        serial_number = adv.serial_number
        devices_by_serial[serial_number] = device
    return devices_by_serial


def find_all_devices():
    devices = {}

    if asphodel.nativelib.usb_devices_supported:
        devices.update(find_usb_devices())

    if asphodel.nativelib.tcp_devices_supported:
        devices.update(find_tcp_devices())

    if not (asphodel.nativelib.usb_devices_supported or
            asphodel.nativelib.tcp_devices_supported):
        # no TCP or USB supported by DLL
        raise Exception("Asphodel library does not support USB or TCP devices")

    return devices


def get_choice(options, prompt=None):
    if prompt is None:
        prompt = "Selection: "

    choices = set(k for k, _m in options)
    choices.discard(None)

    d = {c.lower(): c for c in choices}

    choice = None
    while choice not in d:
        choice = input(prompt)
        if choice:
            choice = choice.lower()

    return d[choice]


def print_options(title, options):
    max_length = max(len(k) for k, _m in options if k is not None)

    print("")
    print("--- {} ---".format(title))
    for key, message in options:
        if key is not None:
            print("{}  {}".format(key.ljust(max_length), message))
        else:
            print("")
            print("{}  {}:".format(" " * max_length, message))
    print("")


def do_string_setting(setting_name, initial_str, default_str):
    print("")
    print("--- {} ---".format(setting_name))
    print("")
    print("Initial setting: {}".format(initial_str))
    print("Default setting: {}".format(default_str))
    value = input("New Value: ")
    return value.encode('UTF-8')


def do_integer_setting(setting_name, initial, default, minimum, maximum):
    if minimum > maximum:
        # they're backwards
        maximum, minimum = minimum, maximum

    print("")
    print("--- {} ---".format(setting_name))
    print("")
    print("Initial setting: {}".format(initial))
    if default is not None:
        print("Default setting: {}".format(default))
    else:
        print("Default setting: invalid")
    prompt = "New Value (range {} to {}): ".format(minimum, maximum)

    while True:
        value_str = input(prompt)
        if value_str:
            try:
                value = int(value_str)
                if minimum <= value <= maximum:
                    return value
            except ValueError:
                continue


def do_float_setting(setting_name, initial, default, minimum, maximum):
    if minimum > maximum:
        # they're backwards
        maximum, minimum = minimum, maximum

    print("")
    print("--- {} ---".format(setting_name))
    print("")
    print("Initial setting: {}".format(initial))
    if default is not None:
        print("Default setting: {}".format(default))
    else:
        print("Default setting: invalid")
    prompt = "New Value (range {} to {}): ".format(minimum, maximum)

    while True:
        value_str = input(prompt)
        if value_str:
            try:
                value = float(value_str)
                if math.isnan(value):
                    # can't compare to min and max
                    return value
                if minimum <= value <= maximum:
                    return value
            except ValueError:
                continue
    return value


def do_choice_setting(setting_name, initial, default, options):
    print_options(setting_name, options)

    for key, name in options:
        if key == str(initial):
            print("Initial setting: {} ({})".format(name, key))
            break
    else:
        print("Initial setting: unknown ({})".format(initial))

    if default is not None:
        for key, name in options:
            if key == str(default):
                print("Default setting: {} ({})".format(name, key))
                break
        else:
            print("Default setting: unknown ({})".format(default))
    else:
        print("Default setting: invalid")

    choice = get_choice(options, "New Value: ")
    return int(choice)


def parse_byte_setting(nvm, setting, default_bytes):
    s = setting.u.byte_setting
    byte_offset = s.nvm_word * 4 + s.nvm_word_byte
    initial = struct.unpack_from(">B", nvm, byte_offset)[0]
    if len(default_bytes) == 1:
        default = default_bytes[0]
    else:
        default = None
    return initial, default


def write_byte_setting(nvm, setting, value):
    s = setting.u.byte_setting
    byte_offset = s.nvm_word * 4 + s.nvm_word_byte
    struct.pack_into(">B", nvm, byte_offset, value)


def do_setting(nvm, setting, custom_enums):
    length = setting.default_bytes_length
    default_bytes = bytes(setting.default_bytes[0:length])
    setting_name = setting.name.decode("UTF-8")

    if setting.setting_type == asphodel.SETTING_TYPE_BYTE:
        initial, default = parse_byte_setting(nvm, setting, default_bytes)
        value = do_integer_setting(setting_name, initial, default, 0, 255)
        write_byte_setting(nvm, setting, value)
    elif setting.setting_type == asphodel.SETTING_TYPE_BOOLEAN:
        initial, default = parse_byte_setting(nvm, setting, default_bytes)
        options = [("0", "False"),
                   ("1", "True")]
        if initial > 1:
            options.append((str(initial), "unknown"))
        value = do_choice_setting(setting_name, initial, default, options)
        write_byte_setting(nvm, setting, value)
    elif setting.setting_type == asphodel.SETTING_TYPE_UNIT_TYPE:
        initial, default = parse_byte_setting(nvm, setting, default_bytes)
        options = []
        for i, name in enumerate(asphodel.unit_type_names):
            options.append((str(i), name))
        if initial >= len(options):
            options.append((str(initial), "unknown"))
        value = do_choice_setting(setting_name, initial, default, options)
        write_byte_setting(nvm, setting, value)
    elif setting.setting_type == asphodel.SETTING_TYPE_CHANNEL_TYPE:
        initial, default = parse_byte_setting(nvm, setting, default_bytes)
        options = []
        for i, name in enumerate(asphodel.channel_type_names):
            options.append((str(i), name))
        if initial >= len(options):
            options.append((str(initial), "unknown"))
        value = do_choice_setting(setting_name, initial, default, options)
        write_byte_setting(nvm, setting, value)
    elif setting.setting_type == asphodel.SETTING_TYPE_STRING:
        s = setting.u.string_setting
        fmt = ">{}s".format(s.maximum_length)
        raw = struct.unpack_from(fmt, nvm, s.nvm_word * 4)[0]
        raw = raw.split(b'\x00', 1)[0]
        raw = raw.split(b'\xff', 1)[0]
        try:
            initial_str = raw.decode("UTF-8")
        except UnicodeDecodeError:
            initial_str = "<ERROR>"
        try:
            default_str = default_bytes.decode("UTF-8")
        except UnicodeDecodeError:
            default_str = "unknown"
        value = do_string_setting(setting_name, initial_str, default_str)
        struct.pack_into(fmt, nvm, s.nvm_word * 4, value)
    elif setting.setting_type == asphodel.SETTING_TYPE_INT32:
        s = setting.u.int32_setting
        initial = struct.unpack_from(">i", nvm, s.nvm_word * 4)[0]
        if len(default_bytes) == 4:
            default = struct.unpack_from(">i", default_bytes, 0)[0]
        else:
            default = None
        value = do_integer_setting(setting_name, initial, default,
                                   s.minimum, s.maximum)
        struct.pack_into(">i", nvm, s.nvm_word * 4, value)
    elif setting.setting_type == asphodel.SETTING_TYPE_INT32_SCALED:
        s = setting.u.int32_scaled_setting
        scaled_min = s.minimum * s.scale + s.offset
        scaled_max = s.maximum * s.scale + s.offset
        initial = struct.unpack_from(">i", nvm, s.nvm_word * 4)[0]
        initial = initial * s.scale + s.offset
        if len(default_bytes) == 4:
            default = struct.unpack_from(">i", default_bytes, 0)[0]
            default = default * s.scale + s.offset
        else:
            default = None
        value = do_float_setting(setting_name, initial, default,
                                 scaled_min, scaled_max)
        unscaled_value = int(round((value - s.offset) / s.scale))
        unscaled_value = max(unscaled_value, s.minimum)
        unscaled_value = min(unscaled_value, s.maximum)
        struct.pack_into(">i", nvm, s.nvm_word * 4, unscaled_value)
    elif setting.setting_type == asphodel.SETTING_TYPE_FLOAT:
        s = setting.u.float_setting
        scaled_min = s.minimum * s.scale + s.offset
        scaled_max = s.maximum * s.scale + s.offset
        initial = struct.unpack_from(">f", nvm, s.nvm_word * 4)[0]
        initial = initial * s.scale + s.offset
        if len(default_bytes) == 4:
            default = struct.unpack_from(">f", default_bytes, 0)[0]
            default = default * s.scale + s.offset
        else:
            default = None
        value = do_float_setting(setting_name, initial, default,
                                 scaled_min, scaled_max)
        unscaled_value = (value - s.offset) / s.scale
        struct.pack_into(">f", nvm, s.nvm_word * 4, unscaled_value)
    elif setting.setting_type == asphodel.SETTING_TYPE_CUSTOM_ENUM:
        s = setting.u.custom_enum_setting
        byte_offset = s.nvm_word * 4 + s.nvm_word_byte
        initial = struct.unpack_from(">B", nvm, byte_offset)[0]
        if len(default_bytes) == 1:
            default = default_bytes[0]
        else:
            default = None
        if s.custom_enum_index >= len(custom_enums):
            # invalid index
            value = do_integer_setting(setting_name, initial, default, 0, 255)
        else:
            options = []
            for i, name in enumerate(custom_enums[s.custom_enum_index]):
                options.append((str(i), name))
            if initial >= len(options):
                options.append((str(initial), "unknown"))
            value = do_choice_setting(setting_name, initial, default, options)
        struct.pack_into(">B", nvm, byte_offset, value)
    else:
        # Note SETTING_TYPE_BYTE_ARRAY and SETTING_TYPE_FLOAT_ARRAY are not
        # supported by this utility as they're not actually used in any devices
        # at the time of this writing
        print("Unsupported setting type!")


def reset_and_reconnect(device):
    device.reset()

    # give the device some time to process
    time.sleep(0.5)

    # reconnect
    start_time = time.monotonic()
    while True:
        try:
            device.reconnect_device(reopen=True)
            break
        except asphodel.AsphodelError:
            if time.monotonic() <= start_time + 5.0:
                time.sleep(0.05)
                continue
            else:
                raise

    device.wait_for_connect(5000)


def do_device_menu(device):
    device.open()
    sn = device.get_serial_number()
    title = "{} Menu".format(sn)

    setting_count = device.get_setting_count()
    settings = [device.get_setting(i) for i in range(setting_count)]
    unassigned_setting_ids = set(range(setting_count))

    setting_category_count = device.get_setting_category_count()
    setting_categories = []
    for i in range(setting_category_count):
        name = device.get_setting_category_name(i)
        category_settings = device.get_setting_category_settings(i)
        setting_categories.append((name, category_settings))
        for setting_id in category_settings:
            unassigned_setting_ids.discard(setting_id)

    custom_enum_counts = device.get_custom_enum_counts()
    custom_enums = {}
    for i, count in enumerate(custom_enum_counts):
        custom_enums[i] = [device.get_custom_enum_value_name(i, v)
                           for v in range(count)]

    nvm_size = device.get_nvm_size()
    nvm = bytearray(device.read_nvm_section(0, nvm_size))

    options = [('p', "Print NVM"),
               ('w', "Write NVM and reset device"),
               ('a', "Abort without saving")]

    if (unassigned_setting_ids):
        options.append((None, "Device Settings"))
        for setting_id in sorted(unassigned_setting_ids):
            setting_name = settings[setting_id].name.decode("UTF-8")
            options.append((str(setting_id), setting_name))
    for category_name, category_settings in setting_categories:
        options.append((None, category_name))
        for setting_id in category_settings:
            setting_name = settings[setting_id].name.decode("UTF-8")
            options.append((str(setting_id), setting_name))

    while True:
        print_options(title, options)
        choice = get_choice(options)

        if choice == 'p':
            print("")
            for line in asphodel.format_nvm_data(nvm):
                print(line)
        elif choice == 'a':
            device.close()
            return
        elif choice == 'w':
            device.erase_nvm()
            device.write_nvm_section(0, nvm)
            reset_and_reconnect(device)
            device.close()
            return
        else:
            setting_id = int(choice)
            do_setting(nvm, settings[setting_id], custom_enums)


def do_main_menu(devices):
    while True:
        options = [('r', "Rescan devices"),
                   ('q', "Quit")]

        if not devices:
            options.append((None, "No Devices"))
        else:
            options.append((None, "Devices"))
            for sn in devices.keys():
                options.append((sn, "Edit device {}".format(sn)))

        print_options("Main Menu", options)
        choice = get_choice(options)

        if choice == 'r':
            devices = find_all_devices()
        elif choice == 'q':
            return
        else:
            do_device_menu(devices[choice])


def main():
    devices = find_all_devices()
    do_main_menu(devices)


if __name__ == "__main__":
    main()



import time

import numpy

import asphodel


def stream_fixed_duration(device, cached_device=None, stream_ids=None,
                          duration=2.0):

    # tighter requirements than usual because this function will be used to
    # verify stream rates
    response_time = 0.005
    buffer_time = 0.05
    timeout = 1000

    device_data = {'unknown_ids': 0,
                   'streams': {}}

    if cached_device is None:
        cached_device = device

    stream_count, filler_bits, id_bits = cached_device.get_stream_count()

    if stream_ids is None:
        # use all streams on the device
        stream_ids = list(range(stream_count))

    max_warm_up = 0.0

    # prepare the info list
    info_list = []
    streams = []
    for stream_id in stream_ids:
        stream_struct = cached_device.get_stream(stream_id)
        streams.append(stream_struct)

        max_warm_up = max(max_warm_up, stream_struct.warm_up_delay)

        channel_info_list = []
        channels = stream_struct.channel_count
        indexes = stream_struct.channel_index_list[0:channels]
        for ch_index in indexes:
            channel_info_list.append(cached_device.get_channel(ch_index))
        info_list.append((stream_id, stream_struct, channel_info_list))

    # create the device decoder
    device_decoder = device.lib.create_device_decoder(info_list, filler_bits,
                                                      id_bits)

    # create & register unknown id callback
    def unknown_id_callback(lost_id):
        device_data['unknown_ids'] += 1
    device_decoder.set_unknown_id_callback(unknown_id_callback)

    # create & register lost packet callback
    for stream_decoder, stream_id in zip(device_decoder.decoders,
                                         device_decoder.stream_ids):
        stream_data = {'lost_packets': 0,
                       'stream': stream_decoder.stream_info,
                       'stream_decoder': stream_decoder,
                       'channels': {},
                       'first_packet_time': None,
                       'last_packet_time': None}
        device_data['streams'][stream_id] = stream_data

        def lost_packet_callback(current, last, stream_data=stream_data):
            lost = (current - last - 1) & 0xFFFFFFFFFFFFFFFF
            stream_data['lost_packets'] += lost

        stream_decoder.set_lost_packet_callback(lost_packet_callback)

        count = stream_decoder.channels
        channel_ids = stream_decoder.stream_info.channel_index_list[0:count]
        for channel_decoder, channel_id in zip(stream_decoder.decoders,
                                               channel_ids):
            channel_data = {'data': [],
                            'channel': channel_decoder.channel_info,
                            'channel_decoder': channel_decoder}
            stream_data['channels'][channel_id] = channel_data

            def channel_callback(counter, data, samples, subchannels,
                                 channel_data=channel_data,
                                 stream_data=stream_data):
                now = time.time()
                d = numpy.array(data).reshape(samples, subchannels)
                channel_data['data'].append((now, counter, d))

                if stream_data['first_packet_time'] is None:
                    stream_data['first_packet_time'] = now
                stream_data['last_packet_time'] = now

            channel_decoder.set_callback(channel_callback)

    stream_counts = cached_device.lib.get_streaming_counts(
        streams, response_time, buffer_time, timeout)

    # warm up the streams
    for stream_id in stream_ids:
        device.warm_up_stream(stream_id, True)

    time.sleep(max_warm_up)

    # enable the streams
    for stream_id in stream_ids:
        device.enable_stream(stream_id, True)

        stream_data = device_data['streams'][stream_id]
        stream_data['start_time'] = time.time()

        # disable warm up so we don't have to worry about it later
        device.warm_up_stream(stream_id, False)

    error_code = 0  # OK
    start_time = None

    def packet_callback(status, stream_packets):
        if status != 0:
            nonlocal error_code
            error_code = status
        else:
            nonlocal start_time
            if start_time is None:
                start_time = time.monotonic()
            for packet in stream_packets:
                device_decoder.decode(packet)

    # enable the stream
    device.start_streaming_packets(*stream_counts, callback=packet_callback)

    while error_code == 0:
        if start_time is not None:
            if time.monotonic() >= start_time + duration:
                break
        device.poll_device(100)

    # disable the streams
    for stream_id in stream_ids:
        device.enable_stream(stream_id, False)

    device.stop_streaming_packets()

    # do a final poll after stopping streaming to clean up
    device.poll_device(0)

    # make sure no error code was set
    asphodel.nativelib.asphodel_error_check(error_code)

    return device_data


def filter_stream_data(device_data, stream_id):
    stream_data = device_data['streams'][stream_id].copy()
    stream_data.update({k: v for k, v in device_data.items()
                        if k != "streams" and not k.startswith("_")})
    return stream_data


def filter_channel_data(device_data, stream_id, channel_id):
    stream_data = filter_stream_data(device_data, stream_id)
    channel_data = stream_data['channels'][channel_id].copy()
    channel_data.update({k: v for k, v in stream_data.items()
                         if k != 'channels' and not k.startswith("_")})
    return channel_data


def unpack_streaming_data(data):
    last_index = None
    indexes = []
    chunks = []

    for _timestamp, index, chunk in data:
        if last_index is not None:
            next_index = last_index + 1
            if index != next_index:
                # create a short nan array to insert into the chunks
                nans = [numpy.nan] * chunk.shape[-1]
                nan_chunk = numpy.array(nans, ndmin=2)
                indexes.append(numpy.array(next_index, ndmin=1))
                chunks.append(nan_chunk)
        last_index = index

        samples = chunk.shape[0]
        indexes.append(numpy.linspace(index, index + 1, num=samples,
                                      endpoint=False))
        chunks.append(chunk)

    if indexes:
        index_array = numpy.concatenate(indexes)
    else:
        index_array = numpy.array([])

    if chunks:
        data_array = numpy.concatenate(chunks)
    else:
        data_array = numpy.array([])

    return (index_array, data_array)


def get_average_measurement(device_data, stream_id, channel_id):
    stream_data = device_data['streams'][stream_id]
    channel_data = stream_data['channels'][channel_id]
    unpacked_data = unpack_streaming_data(channel_data['data'])
    values = unpacked_data[1]
    return numpy.nanmean(values, axis=0)


class StreamingCache:
    caches = {}  # key=cache_device, value=device_data

    def __init__(self, cached_device, stream=None, stream_id=None,
                 channel=None, channel_id=None):
        # NOTE: the cached device shouldn't be used for any streaming
        self.cached_device = cached_device
        self.stream = stream
        self.stream_id = stream_id
        self.channel = channel
        self.channel_id = channel_id

    def __call__(self, device):
        device_data = self.caches.get(self.cached_device, None)

        if device_data is None:
            # get new data
            device_data = stream_fixed_duration(device, self.cached_device)

            # clean up after; formerly part of stream_fixed_duration()
            device.flush()

            # save it to the cache
            self.caches[self.cached_device] = device_data

        # see if the stream data is still valid
        stale_objs = device_data.setdefault('_stale_objs', set())
        if self in stale_objs:
            # get new data
            device_data = stream_fixed_duration(device, self.cached_device)

            # clean up after; formerly part of stream_fixed_duration()
            device.flush()

            stale_objs = set()
            device_data['_stale_objs'] = stale_objs

            # save it to the cache
            self.caches[self.cached_device] = device_data

        # mark the current object as stale, so the next call will give new data
        stale_objs.add(self)

        if self.stream_id is None:
            data = device_data
        elif self.channel_id is None:
            data = filter_stream_data(device_data, self.stream_id)
        else:
            data = filter_channel_data(device_data, self.stream_id,
                                       self.channel_id)

        return data
